/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2009. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("jazz.ui.SearchBox");

dojo.require("dojo.i18n");
dojo.require("dojo.string");

dojo.require("dijit._Templated");
dojo.require("dijit._Widget");
dojo.require("dijit.Menu");

dojo.requireLocalization("jazz.ui", "SearchBoxMessages");

(function() {

var _DEFAULT_DELAY = 500;
var _DEFAULT_TEXT = "";

var _messages = dojo.i18n.getLocalization("jazz.ui", "SearchBoxMessages");

dojo.declare("jazz.ui.SearchBox", [dijit._Widget, dijit._Templated], {
	
	// summary:
	//		A customizable search box with a text field, search button, and scope
	//		selection menu.
	//
	//      NOTE: This is provisional API. Clients using this API do so with the
	//		awareness that this class may be removed or substantially changed at
	//		future times without warning.
	// description:
	//		Provides a number of parameters in the constructor for the component-
	//		specific behaviors, such as updating the results, performing the search,
	//		search scopes, etc. There are two modes of searching:
	//			1) Search as you type - A results drop down will appear below the
	//				search results and will be updated as you type. This feature is
	//				enabled by supplying an onUpdate function in the constructor.
	//			2) Traditional - Type in the query and press enter or click on the search
	//				button. Enabled by providing an onExecute function in the constructor.
	//		At least one of onUpdate and onExecute must be supplied. Both are allowed
	//		simultaneously.
	// since:
	//		0.7

	templatePath: dojo.moduleUrl("jazz.ui", "templates/SearchBox.html"),
	
	constructor: function(/* object */ args) {
		// summary:
		//		Constructs a new search box with the given parameters.
		// scopes:
		//		An array of objects containing information about each scope type to
		//		provide for the search box. Each object must have 3 properties:
		//			- id (string): A unique string representing the scope
		//			- label (string): A human-readable label for the scope
		//			- icon (string): A path to a 16x16 image for the scope icon
		//		e.g.:
		//			[
		//				{ id: "scope1", label: "First Scope", icon: dojo.moduleUrl("my.module", "images/scope1.gif") },
		//				{ id: "scope2", label: "Second Scope", icon: dojo.moduleUrl("my.module", "images/scope2.gif") },
		//			]
		// onUpdate:
		//		An optional function that, if specified, will be called when the search
		//		results drop-down (search as you type) needs to be updated. The function will
		//		be given the following parameters:
		//			- query (string): The text to search for
		//			- scope (string): The id of the currently selected scope (see scopes param)
		//			- callback (function): The callback function for returning the search
		//				results. The function takes one parameter, an array of objects
		//				containing the search results, each with a label parameter containing
		//				HTML formatted markup for the result, a handler function (optional)
		//				to be executed when the result is selected, and an array of child
		//				results (optional), which will be grouped under the parent.
		//				e.g.: 
		//					[ 
		//						{ label: "<b>First</b> result", handler: function() {...} }
		//						{ label: "<b>Second</b> result", children: [ {...}, {...}] }
		//					]
		//			- errorCallback (function): The callback to use for reporting errors.
		//				Accepts one string parameter containing the error message.
		//		One of onUpdate, onExecute must be specified.
		// onExecute:
		//		An optional function that, if specified, will be called when the user clicks
		// 		on the search button or presses enter. It accepts the following parameters:
		//			- query (string): The text to search for
		//			- scope (string): The id of the currently selected scope (see scopes param)
		//		If not specified, onUpdate will be called instead, and the first result with
		//		a handler will be executed.
		//		One of onUpdate, onExecute must be specified.
		// delay:
		//		The optional minimum delay in milliseconds between each update (see
		//		onUpdate param). If not specified, a default will be used.
		// tooltip:
		//		The tooltip (string) to show for the search field on mouse hover.
		// defaultText:
		//		Optional default text (string) to show in the search box when empty, e.g.
		//		"Type to search".
		// noResultsLabel:
		//		An optional HTML-formatted label (string) to show in the update results
		//		when no results were found, e.g. "No results found".
		// minResultsWidth:
		//		An optional minimum width (int) in pixels for the drop down results shown
		//		as you type.
		this._messages = _messages;
		this._scopes = args.scopes;
		this._onUpdate = args.onUpdate;
		this._onExecute = args.onExecute;
		this._delay = args.delay ? args.delay : _DEFAULT_DELAY;
		this._tooltip = args.tooltip;
		this._defaultText = args.defaultText ? args.defaultText : _DEFAULT_TEXT;
		this._noResultsLabel = args.noResultsLabel ? args.noResultsLabel : _messages.noResultsLabel;
		this._minResultsWidth = args.minResultsWidth;
		
		// to skip stale responses that come out of order
		this._sequenceCounter = 0;
		this._lastResponseReceived = -1;
		
		this._isHoveringSearchButton = false;
	},
	
	postCreate: function() {
		this._searchText.value = this._defaultText;
		this._searchText.title = this._tooltip;
		
		this._searchButtonSpan.appendChild(document.createTextNode(this._tooltip));
		
		this._setScope(this._scopes[0], false);
		this.connect(this._searchText, "onkeyup", this._onchange);
		this.connect(this._searchText, "oninput", this._onchange);
		this.connect(this._searchText, "onkeypress", this._onkeypress);
		this.connect(this._searchText, "onfocus", this._onfocus);
		this.connect(this._searchText, "onblur", this._onblur);
		this.connect(this._scopeMenu, "onclick", this._openMenu);
		this.connect(this._scopeMenu, "onkeypress",this._sniffKey);
		this.connect(this._searchButton, "onclick",this._onclickSearchButton);
		this.connect(this._searchButton, "onmouseover",this._onmouseoverSearchButton);
		this.connect(this._searchButton, "onmouseout",this._onmouseoutSearchButton);
	},
	
	destroy: function() {
		if (this._results) {
			this._results.destroyRecursive();
			delete this._results;
		}
		if (this._menu) {
			this._menu.destroyRecursive();
			delete this._menu;
		}
		this.inherited(arguments);
	},

	_onchange: function(e) {
		if (e && e.keyCode != dojo.keys.ENTER) {
			var query = dojo.trim(this._searchText.value);		
			if (query != this._currentQuery) {
				this._currentQuery = query;
				if (!this._timeout) {
					// wait a bit then check
					this._timeout = setTimeout(dojo.hitch(this, this._update, false), 100);
				}
			}
			this._updateSearchButton();
		}
	},
	
	_onkeypress: function(e) {
		if (e && e.keyCode == dojo.keys.DOWN_ARROW || e.keyCode == dojo.keys.ENTER) {
			if (e.keyCode == dojo.keys.DOWN_ARROW && this._results) {
				// down arrow from text field jumps to drop down
				this._results.focus();
			}
			if (e.keyCode == dojo.keys.ENTER) {
				this._performExecute(true);
			}
			dojo.stopEvent(e);
		}
		if (e && e.keyCode == dojo.keys.TAB && !e.shiftKey && this._isResultsOpen) {
			this._results.focus();
			dojo.stopEvent(e);
		}
	},
	
	_onfocus: function(e) {
		if (this._searchText.value == this._defaultText) {
			this._searchText.value = "";
		}
		dojo.removeClass(this._searchText, "inactive");
		this._closeMenu();
	},

	_onblur: function(e) {
		// kill the periodic update
		if (this._timeout) {
			clearTimeout(this._timeout);
			delete this._timeout;
		}
		// show default help text if empty
		if (dojo.trim(this._searchText.value) == "") {
			dojo.addClass(this._searchText, "inactive");
			this._searchText.value = this._defaultText;
		}
	},
	
	_onclickSearchButton: function(e) {
		if (e) {
			dojo.stopEvent(e);
		}
		this._performExecute(false);
	},
	
	_onmouseoverSearchButton: function() {
		this._isHoveringSearchButton = true;
		this._updateSearchButton();
	},
	
	_onmouseoutSearchButton: function() {
		this._isHoveringSearchButton = false;
		this._updateSearchButton();
	},
	
	_performExecute: function(/* boolean */ performDefault) {
		if (this._currentQuery && this._currentQuery.length > 0) {
			if (this._onExecute) {
				// stop updating
				if (this._timeout) {
					clearTimeout(this._timeout);
					delete this._timeout;
				}
				this._closeResults();
				this._onExecute(this._currentQuery, this._scope.id);
			}
			// no global execute handler
			else if (this._onUpdate) {
				if (performDefault && this._isResultsOpen) {
					// if open and up to date, execute first handler
					if (this._lastResponseReceived == this._sequenceCounter) {
						if (this._results.performDefault()) {
							// close results and ignore any old pending searches
							this._lastResponseReceived = this._sequenceCounter;
							this._closeResults();
						}
					}
				}
				// force show results
				else {
					this._update(true);
				}
			}
		}
	},
	
	_showError: function(/* string */ errorMsg) {
		this._showResults({message: errorMsg}, this._minResultsWidth);
	},
	
	_showResults: function(/* object[] */ results, /* number */ minWidth) {
		if (!results || results.length == 0) {
			results = [{label: this._noResultsLabel}];
		}
		if (!this._results) {
			this._results = new net.jazz.web.ui.internal.SearchResults({predecessor: this._searchText, successor: this.domNode.nextSibling, onSelect: dojo.hitch(this, this._closeResults)});
			this._resultsCloseOnBody = this.connect(this, "_onBlur", this._closeResults);
		}
		
		var pos = dojo.coords(this._searchWrapper);
		pos.x += document.documentElement.scrollLeft;
		pos.y += document.documentElement.scrollTop;
		
		// use client specified min width if available, else size according to search box
		var width = minWidth;
		if (!width || width < pos.w) {
			width = pos.w;
		}
		
		this._results.setContent(results, width);
		
		// open if needed
		if (!this._isResultsOpen && this._currentQuery.length > 0) {
			/*
			 * Workaround to avoid jumping. Can't use node style because dijit uses it,
			 * so use a class instead.
			 */
			dojo.addClass(this._results.domNode, "net-jazz-web-ui-internal-SearchResults-hidden");
			dijit.popup.open({ popup: this._results, parent: this, x: pos.x, y: pos.y + pos.h - 1, onExecute: dojo.hitch(this,this._closeResults), onCancel: dojo.hitch(this,this._closeResults), orient: 'BL' });
			dojo.removeClass(this._results.domNode, "net-jazz-web-ui-internal-SearchResults-hidden");
			this._isResultsOpen = true;
		}
	},
	
	_closeResults: function(e){
		if (e) {
			dojo.stopEvent(e);
		}
		if (this._results) {
			dijit.popup.close(this._results);
			this._results.destroyRecursive();
			delete this._results;
			this._isResultsOpen = false;
		}
	},
	
	_update: function(/* boolean */ force) {
		if (this._onUpdate && ((this._currentQuery != this._lastQuery) || force)) {
			this._lastQuery = this._currentQuery;
			if (this._currentQuery.length > 0) {
				if (!this._isResultsOpen) {
					this._showResults([{label: _messages.loading}]);
				}
				/*
				 * Use the sequence counter to make sure we never show an old
				 * stale response when we've already received a more recent one.
				 */
				var callbackObj = {
					_self: this,
					_sequenceCounter: ++this._sequenceCounter,
					callback: function(/* object[] */ results) {
						// only show responses newer than the previous; ignore older ones
						if (this._sequenceCounter > this._self._lastResponseReceived) {
							this._self._lastResponseReceived = this._sequenceCounter;
							this._self._showResults(results, this._self._minResultsWidth);
						}
					}
				};
				this._onUpdate(this._currentQuery, this._scope.id, dojo.hitch(callbackObj, callbackObj.callback), dojo.hitch(this, this._showError));
			}
			else {
				// ignore any pending searches by syncing up the counters
				this._lastResponseReceived = this._sequenceCounter;
				this._closeResults();
			}
			this._timeout = setTimeout(dojo.hitch(this, this._update, false), this._delay);
		}
		else {
			delete this._timeout;
		}
	},
	
	_updateSearchButton: function() {
		var hasText = (this._currentQuery && this._currentQuery.length > 0);
		this._searchButton.className = "SearchButton " + (hasText ? "SearchButtonEnabled" : "SearchButtonDisabled");
	},
	
	_sniffKey: function(e){
		if (e.keyCode == dojo.keys.DOWN_ARROW || e.keyCode == dojo.keys.ENTER) {
			this._openMenu();
			dojo.stopEvent(e);
		}
	},
	
	_openMenu: function(e) {
		if (e){
			dojo.stopEvent(e);
		}
		this._closeResults();
		if (this._menu) {
			this._closeMenu();
			return;
		}
		
		var menu = new dijit.Menu();
		dojo.addClass(menu.domNode, "jazz-ui-SearchBox-ScopeMenu");
		var scopes = this._scopes;
		for (var i=0;i<scopes.length;i++) {
			var label = "<img class=\"itemIcon\" src=\"" + scopes[i].icon + "\"><span class=\"itemLabel\">" + scopes[i].label + "</span>";
			var args = {label: label, onClick: dojo.hitch(this, this._setScope, scopes[i], true)};
			if(scopes[i] === this._scope) {
				args.iconClass = "selectedItem";
			}
			menu.addChild(new dijit.MenuItem(args));
		}
		
		var pos = dojo.coords(this._scopeMenu);
		pos.x += document.documentElement.scrollLeft;
		pos.y += document.documentElement.scrollTop;

		dojo.addClass(this._scopeMenu, "ScopeMenu-selected");
		dijit.popup.open({ popup: menu, parent: this, x: pos.x - (dojo.isFF ? 0 : 1), y: pos.y + pos.h - (dojo.isFF ? 1 : 2), onExecute: dojo.hitch(this,this._closeMenu), onCancel: dojo.hitch(this,this._closeMenu), orient: 'BL' });
		
		//register event handler that causes menu to close
		this._menuCloseOnBody = this.connect(this,"_onBlur",this._closeMenu);
		this._menuKeyHandler = this.connect(menu.domNode,"onkeypress",this._onMenuKeyPress);
		this._menu = menu;
		menu.focus();
	},
	
	_closeMenu: function(e){
		if (e) {
			dojo.stopEvent(e);
		}
		if (this._menu) {
			dojo.removeClass(this._scopeMenu, "ScopeMenu-selected");
			this.disconnect(this._menuCloseOnBody);
			this.disconnect(this._menuKeyHandler);
			dijit.popup.close(this._menu);
			this._menu.destroyRecursive();
			this._menu = null;
		}
	},
	
	_onMenuKeyPress: function(e) {
		if (e.keyCode == dojo.keys.TAB) {
			if (e.shiftKey) {
				// previous item is menu drop down
				this._scopeMenu.focus();
			}
			else {
				// next item is text field
				this._searchText.focus();
			}
		}
	},
	
	_setScope: function(/* object */ scope, /* boolean */ focusText) {
		this._scope = scope;
		this._scopeMenu.title = dojo.string.substitute(_messages.tooltip, [ this._scope.label ]);
		var scopes = this._scopes;
		this._menuImg.src = this._scope.icon;
		if (focusText) {
			this._searchText.focus();
		}
		// update if already showing
		if (this._isResultsOpen) {
			this._update(true);
		}
	}
});

/*
 * The drop-down search result widget.
 */
dojo.declare("net.jazz.web.ui.internal.SearchResults",
	[dijit._Widget, dijit._Templated], {
	
	templateString: "<div class=\"net-jazz-web-ui-internal-SearchResults\"></div>",
	
	constructor: function(args) {
		this._predecessor = args.predecessor;
		this._successor = args.successor;
		this._onSelect = args.onSelect;
	},
	
	focus: function() {
		this._menu.focus();
	},
	
	performDefault: function() {
		if (this._defaultHandler) {
			this._defaultHandler();
			return true;
		}
		return false;
	},
	
	setContent: function(results, width) {
		delete this._defaultHandler;
		this._results = results;
		this._width = width;

		// create a new menu
		var menu = new net.jazz.web.ui.internal.CustomMenu({predecessor: this._predecessor, successor: this._successor});
		for (var i=0;i<results.length;i++) {
			this._addItem(menu, results[i], 0);
			if (!this._defaultHandler && results[i].handler) {
				this._defaultHandler = results[i].handler;
			}
		}
		
		// create container
		var content = document.createElement("div");
		content.className = "content";
		content.style.width = (this._width - 2) + "px";
		content.appendChild(menu.domNode);
		
		// swap in new container
		if (this._content) {
			this.domNode.replaceChild(content, this._content);
		}
		else {
			this.domNode.appendChild(content);
		}
		this._content = content;
		
		// destroy old menu
		if (this._menu) {
			this._menu.destroyRecursive();
		}
		this._menu = menu;
	},
	
	_addItem: function(menu, result, indent) {
		var buf = ["<div class=\"entry\" style=\"padding-left: "];
		buf.push((indent*1.25));
		buf.push("em\">");
		buf.push(result.label);
		buf.push("</div>");
		if (result.handler) {
			var handlerObj = {
				_self: this,
				_result: result,
				handle: function() {
					this._self._onSelect();
					this._result.handler();
				}
			};
			var item = new dijit.MenuItem({label: buf.join(""), onClick: dojo.hitch(handlerObj, handlerObj.handle)});
		}
		else {
			var item = new net.jazz.web.ui.internal.DisabledMenuItem({label: buf.join("")});
		}
		menu.addChild(item);
		
		// add children recursively
		var children = result.children;
		if (children) {
			var childIndent = indent + 1;
			for (var i=0;i<children.length;i++) {
				this._addItem(menu, children[i], childIndent);
			}
		}
	}
});

/*
 * A specialized menu that handles up/down keys differently than normal.
 * If you press down arrow and you're on the last node, it stays instead of
 * wrapping. If you're on the first node and you press up arrow, it will leave
 * the menu and go to a designated predecessor node.
 */
dojo.declare("net.jazz.web.ui.internal.CustomMenu",
	dijit.Menu, {
	
	/*
	 * Constructs the menu with the given predecessor node that will receive
	 * focus when you press up arrow on the first node.
	 */
	constructor: function(args) {
		this._predecessor = args.predecessor;
		this._successor = args.successor;
	},
	
	postCreate: function() {
		this.inherited(arguments);
		this.connect(this, "onkeypress", this._onKeyPress);
	},
	
	/*
	 * Override to prevent wrapping on the last node.
	 */
	focusNext: function() {
		if (this.focusedChild) {
			var children = this.getChildren();
			if (this.focusedChild == children[children.length-1]) {
				// prevent wrapping
				return;
			}
		}
		this.inherited(arguments);
	},

	/*
	 * Override to jump to predecessor node if on first node.
	 */	
	focusPrev: function() {
		if (this.focusedChild) {
			var children = this.getChildren();
			if (this.focusedChild == children[0]) {
				// jump to the menu's predecessor
				this._onChildBlur(this.focusedChild);
				this.focusedChild = null;
				this._predecessor.focus();
				return;
			}
		}
		this.inherited(arguments);
	},
	
	onItemHover: function(/* MenuItem */ item) {
		dojo.addClass(item.domNode, "dijitMenuItemHover");
	},
	
	onItemUnhover: function(/* MenuItem */ item) {
		dojo.removeClass(item.domNode, "dijitMenuItemHover");
	},
	
	_onKeyPress: function(e) {
		if (e.keyCode == dojo.keys.TAB) {
			if (e.shiftKey) {
				this._predecessor.focus();
			}
			else {
				this._successor.focus();
			}
		}
	}
});

/*
 * A grayed and non-functional menu item. You can't click or highlight it, and
 * up/down arrow keys will skip it.
 */
dojo.declare("net.jazz.web.ui.internal.DisabledMenuItem",
	dijit.MenuItem, {

	/*
	 * Override to disable the menu item.
	 */
	postCreate: function(){
		this.inherited(arguments);
		dojo.addClass(this.domNode, "disabled");
		dojo.setSelectable(this.domNode, false);
	},
	
	/*
	 * Never receives focus.
	 */
	isFocusable: function(){
		return false;
	}
});

})();