/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2009. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("jazz.ui.tree.AbstractItem");

dojo.require("dijit._Widget");

(function() {

dojo.declare("jazz.ui.tree.AbstractItem", dijit._Widget, {
	// summary:
	//		The base item that all items for use in jazz.ui.Tree and jazz.ui.ModeledTree
	//		must subclass.
	//
	//      NOTE: This is provisional API. Clients using this API do so with the
	//		awareness that this class may be removed or substantially changed at
	//		future times without warning.
	// description:
	//		This class must be subclassed for use in the tree as it does not provide the ui.
	// since:
	//		0.7
	
	// PUBLIC API
	
	tree: null,	// READ ONLY. A reference to the tree this item is in.
	
	update: function() {
		// summary:
		//		Called when the item should be updated.
		// description:
		//		When the tree (or part of it) is refreshed, this function will be called. It
		//		can be used to allow the item to update itself in the even that its parameters
		//		have changed. Subclasses should override as required.
	},
	
	getFocusNode: function() {
		// summary:
		//		Called by the tree to get the dom node that accepts browser focus.
		// description:
		//		When the tree needs to set browser focus to the item, it will call this function
		//		to get the appropriate dom node to use. The default implementation returns the
		//		widget's root node (assuming the subclass populates this.domNode through
		//		templating or some other means).
		return this.domNode;
	},
	
	getId: function() {
		// summary:
		//		Returns the unique id of the item.
		// description:
		//		This function returns a unique id for the tree item. The default implementation
		//		uses the dojo widget id. This is guaranteed to be unique, but may be different
		//		on every load. Several functions of jazz.ui.ModeledTree require that its tree
		//		items return consitent ids based on the model objects. These functions include:
		//			-ModeledTree.expandTo
		//			-ModeledTree.getState
		//			-ModeledTree.restoreState
		return this.id;
	},
	
	// FUNCTIONS THAT CAN BE CALLED BUT NOT OVERRIDEN
	
	addChild: function(item, index) {
		// summary:
		//		Adds a new item as a child of this item.
		// description:
		//		Adds the provided item as a child of this item. If the index is specified, the item
		//		will be added in the desired position relative to the rest of the children.
		// item:
		//		A subclass of jazz.ui.tree.AbstractItem that represents the node to be added.
		// index:
		//		The index to insert the item relative to the other children.
		item.tree = this.tree;
		item._parent = this;
		if (!index && index != 0)
			this._children.push(item);
		else
			this._children.splice(index,0,item);
		this._notifyTree();
	},
	
	getChildren: function() {
		// summary:
		//		Returns the children of this item.
		// description:
		//		Returns an array of jazz.ui.tree.AbstractItem instances that represent the children of
		//		this item. The array will be ordered the same as the nodes appear in the tree.
		return this._children;
	},
	
	getDepth: function() {
		// summary:
		//		Returns the depth of the item in the tree.
		// description:
		//		Returns an integer representing the depth of the item in the tree. Roots will return 0,
		//		thier direct children will return 1, etc.
		if (this._parent)
			return this._parent.getDepth() + 1;
		return -1;
	},
	
	getIndex: function(child) {
		// summary:
		//		Returns the index of the indicated child.
		// description:
		//		Returns an integer representing the index of the provided child relative to the other
		//		children of this item.
		// child:
		//		The jazz.ui.tree.AbstractItem instance whose index is required.
		for (var i = 0; i < this._children.length; i++)
			if (child.getId() == this._children[i].getId())
				return i;
	},
	
	getNextSibling: function() {
		// summary:
		//		Returns the next sibling of this item.
		// description:
		//		Returns the tree item immediately following this item under the same parent node. If
		//		this is the last child of the parent, returns null.
		if (!this._parent)
			return null;
		var i = this._parent.getIndex(this);
		var arr = this._parent.getChildren();
		if (i == arr.length - 1)
			return null;
		return this._parent.getChildren()[i+1];
	},
	
	getParent: function() {
		// summary:
		//		Returns the parent of this item.
		// description:
		//		Returns the tree item that this item is a child of. If this is a root of the tree,
		//		null is returned.
		return this.isRoot() ? null : this._parent;
	},
	
	getPreviousSibling: function() {
		// summary:
		//		Returns the previous sibling of this item.
		// description:
		//		Returns the tree item immediately preceeding this item under the same parent node. If
		//		this is the first child of the parent, returns null.
		if (!this._parent)
			return null;
		var i = this._parent.getIndex(this);
		if (i == 0)
			return null;
		return this._parent.getChildren()[i-1];
	},
	
	isExpanded: function() {
		// summary:
		//		Returns a boolean indicating if the item is expanded.
		// description:
		//		If this item is a leaf node, false is returned (leaf nodes cannot be expanded).
		//		Otherwise, the expansion state of the item is returned.
		return !this.isLeaf() &&
			dojo.hasClass(this.tree._branches[this.getId()].root,"open") &&
			(this.isRoot() || this._parent.isExpanded());
	},
	
	isLeaf: function() {
		// summary:
		//		Returns a boolean indicating if the item is a leaf node.
		return this.tree._isLeaf(this);
	},
	
	isRoot: function() {
		// summary:
		//		Returns a boolean indicating if the item is a root of its tree.
		return this._parent.getId() == this.tree._root.getId();
	},
	
	isVisible: function() {
		// summary:
		//		Returns a boolean indicating if the item is currently visible.
		// description:
		//		Will return false if its parent or any of its parent's parents are collapsed. Roots
		//		will always return true.
		return this.isRoot() || this._parent.isExpanded();
	},
	
	refresh: function() {
		// summary:
		//		Refreshes the tree from this item.
		// description:
		//		Refreshes this item and all items below it in the tree.
		this.tree._refreshItem(this);
	},
	
	removeChild: function(child, preserve) {
		// summary:
		//		Removes a child from this item.
		// description:
		//		Removes the indicated child from the item.
		// child:
		//		A reference to the jazz.ui.tree.AbstractItem that should be removed. This must be
		//		a valid child of this item.
		// preserve:
		//		A boolean indicating if the tree item should be preserved. If this is not true,
		//		the removed item will be destroyed. This should be set to true if the item will be
		//		held to add back to the tree later.
		for (var i = 0; i < this._children.length; i++) {
			if (child.getId() == this._children[i].getId()) {
				this._children.splice(i,1);
				if (this.tree)
					this.tree._removeItem(child);
				if (preserve)
					delete child.tree;
				else
					child.destroy();
				break;
			}
		}
	},
	
	setExpanded: function(bool) {
		// summary:
		//		Sets the expanded state of the item.
		// description:
		//		Expands or collapses the item based on the provided boolean. The state of all
		//		children will be maintained in the case of further expansion/collapsing of this item.
		// bool:
		//		A boolean indicating if the node should be expanded or collapsed. Set this to true to
		//		expand the item, or false to collapse.
		this.tree._setExpanded(this, bool);
	},
	
	toggleExpanded: function() {
		// summary:
		//		Toggles the expanded state of the item.
		// description:
		//		Expands or collapses the item based on its previous state. The state of all children
		//		will be maintained in the case of further expansion/collapsing of this item.
		this.setExpanded(!this.isExpanded());
	},
	
	// INTERNAL PROPERTIES AND FUNCTIONS
	
	_parent: null,
	postCreate: function() {
		this._children = [];
	},
	
	_notifyTree: function() {
		if (this.tree)
			this.tree._refreshItem(this);
	}
	
});

})();